<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Stack Visualization</title>
    <style>
    {% include 'partials/stack-style.css' %}
    </style>
</head>
<body>
    {% include 'partials/stack-content.html.j2' %}
</body>
</html>
